"""
Join operations for DataFrame.

This module provides join and set operations that can be mixed into
the DataFrame class to add join capabilities.
"""

from .operations import JoinOperations

__all__ = ["JoinOperations"]
